#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # Traitement amélioré des catégories
# 
# Ce notebook présenté des encoding différents de ceux implémentées dans [scikit-learn](http://scikit-learn.org/stable/).

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# In[2]:


get_ipython().run_line_magic('matplotlib', 'inline')


# On construit un jeu très simple avec deux catégories, une entière, une au format texte.

# In[3]:


import pandas
import numpy
df = pandas.DataFrame(dict(cat_int=[10, 20, 10, 39, 10, 10, numpy.nan],
                          cat_text=['catA', 'catB', 'catA', 'catDD', 'catB', numpy.nan, 'catB']))
df


# ## Une API un peu différente
# 
# Le module [Category Encoders](http://contrib.scikit-learn.org/categorical-encoding/index.html) implémente d'autres options avec une [API](http://contrib.scikit-learn.org/categorical-encoding/onehot.html) un peu différente puisqu'il est possible de spécifier la colonne sur laquelle s'applique l'encoding.

# In[4]:


from category_encoders import OneHotEncoder
OneHotEncoder(cols=['cat_text']).fit_transform(df)


# ## Autres options

# In[5]:


import category_encoders
encoders = []
for k, enc in category_encoders.__dict__.items():
    if 'Encoder' in k:
        encoders.append(enc)
encoders


# In[6]:


dfi = df[['cat_text']].copy()
dfi["copy"] = dfi['cat_text']
for encoder in encoders:
    if 'Leave' in encoder.__name__ or \
       'Target' in encoder.__name__ or \
       'WOE' in encoder.__name__:
        continue
    enc = encoder(cols=['cat_text'])
    try:
        out = enc.fit_transform(dfi)
    except Exception as e:
        print("Issue with '{0}' due to {1}".format(encoder.__name__, e))
        continue
    print('-----', encoder.__name__)
    print(out)
    print('-----')    


# ## Utilisation de la cible
# 
# Certains encoding optimise l'encoding en fonction de la cible à prédire lors d'un apprentissage supervisé. Les deux encoders suivant prédisent la cible en fonction de la catégorie ou essayent d'optimiser l'encoding de la catégorie en fonction de la cible à prédire. En particulier, l'encoder [LeaveOneOut](http://contrib.scikit-learn.org/categorical-encoding/leaveoneout.html) associe à chaque modéalité la moyenne des valeurs observées sur une autre colonne pour chaque ligne associée à cette modalité.

# In[7]:


dfy = df.sort_values('cat_text').reset_index(drop=True).copy()
dfy['cat_text_copy'] = dfy['cat_text']
dfy['y'] = dfy.index * dfy.index + 10
dfy['y_copy'] = dfy.y
dfy


# In[8]:


categories = dfy.drop('y', axis=1)
label = dfy.y
binary_label = label == 10  # dummy one

for encoder in encoders:
    enc = encoder(cols=['cat_text'])
    try:
        out = enc.fit_transform(categories)
    except Exception as e:
        out = pandas.DataFrame()
    try:
        outy = enc.fit_transform(categories, label)
    except ValueError as e:
        if "must be binary" not in str(e):
            continue
        outy = enc.fit_transform(categories, binary_label)
    if not out.equals(outy):
        print('-----', encoder.__name__)
        print(outy)
        print('-----')  


# ## A propos du LeaveOneOut
# 
# Cet encoder ne produit qu'une seule colonne. Dans le cas d'une régression linéaire, la valeur est la moyenne de la cible $y$ sur l'ensemble des lignes associées à cette catégorie. On reprend un exemple déjà utilisé.

# In[9]:


perm = numpy.random.permutation(list(range(10)))
n = 1000
X1 = numpy.random.randint(0, 10, (n,1))
X2 = numpy.array([perm[i] for i in X1])
eps = numpy.random.random((n, 1))
Y = X1 * (-10) - 7 + eps
data = pandas.DataFrame(dict(X1=X1.ravel(), X2=X2.ravel(), Y=Y.ravel()))
data.head()


# In[10]:


from sklearn.model_selection import train_test_split
data_train, data_test = train_test_split(data)


# In[11]:


data_train = data_train.reset_index(drop=True)


# In[12]:


from category_encoders import LeaveOneOutEncoder
le = LeaveOneOutEncoder(cols=['X2'])
X = data_train.drop('Y', axis=1)
le.fit(X, data_train.Y)
data_train2 = le.transform(X)
data_train2.head()


# In[13]:


data_train2 = data_train2.reset_index(drop=True)


# In[14]:


from sklearn.linear_model import LinearRegression
model = LinearRegression()
model.fit(data_train2[["X2"]], data_train['Y'])


# In[15]:


data_test = data_test.reset_index(drop=True)


# In[16]:


from sklearn.metrics import r2_score
data_test2 = le.transform(data_test.drop("Y", axis=1))
r2_score(data_test['Y'], model.predict(data_test2[['X2']]))


# Le coefficient $R^2$ est proche de 1, la régression est quasi parfaite. L'encodeur ``LeaveONeOutEncode`` utilise la cible pour maximiser le coefficient $R^2$ si le modèle utilisé pour prédire est une régression linéaire. Vous trouverez une idée de la démonstration dans cet énoncé : [ENSAE TD noté 2016](http://www.xavierdupre.fr/site2013/enseignements/tdnoteseul/td_note_2017.pdf).

# In[17]: